use std::fmt::Debug;

use binrw::{BinRead, BinWrite};
use diff::Diff;
use modular_bitfield::{
    bitfield,
    specifiers::{B1, B2, B21, B4},
};

use crate::{
    array_of::ArrayOf,
    field_of::FieldOf,
    types::{Field, WString, F64, U32},
};

use super::{line::Lines, Object, ObjectCore};

#[bitfield(bits = 32)]
#[derive(BinRead, BinWrite, Copy, Clone, Debug, Default, Diff, PartialEq)]
#[diff(attr(
    #[derive(Debug, PartialEq)]
))]
#[br(map = Self::from_bytes)]
#[bw(map = |&x| Self::into_bytes(x))]
pub struct ObjectFlags {
    #[skip]
    __: B1,
    pub follow_edge_once: B1,
    #[skip]
    __: B4,
    pub auto_rotate_hatch_angle: B1,
    pub average_distribute_line: B1,
    #[skip]
    __: B2,
    pub cross_patch: B1,
    #[skip]
    __: B21,
}

#[derive(BinRead, BinWrite, Debug, Diff, PartialEq)]
#[diff(attr(
    #[derive(Debug, PartialEq)]
))]
pub struct HatchLine {
    lines: ArrayOf<Lines>,
}

#[derive(BinRead, BinWrite, Diff, PartialEq)]
#[diff(attr(
    #[derive(Debug, PartialEq)]
))]
#[brw(magic(46u32))] // Number of fields in this struct
pub struct HatchSettings1 {
    pub mark_contour: U32,
    _unknown_1: Field,
    pub pen: U32,
    pub flags: FieldOf<ObjectFlags>,
    pub edge_offset: F64,
    pub line_spacing: F64,
    pub start_offset: F64,
    pub end_offset: F64,
    pub angle: F64,
    _unknown_2: [Field; 9],
    pub auto_rotate_angle: F64,
    _unknown_3: [Field; 10],
    pub line_reduction: F64,
    _unknown_4: [Field; 2],
    pub loop_count: U32,
    _unknown_5: [Field; 2],
    pub loop_distance: F64,
    _unknown_6: [Field; 5],
    pub contour_first: U32,
    pub count: U32,
    _unknown_7: [Field; 3],
}

// Custom Debug implementation to only print known fields
impl Debug for HatchSettings1 {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("HatchSettings1")
            .field("mark_contour", &self.mark_contour)
            .field("pen", &self.pen)
            .field("flags", &self.flags)
            .field("edge_offset", &self.edge_offset)
            .field("line_spacing", &self.line_spacing)
            .field("start_offset", &self.start_offset)
            .field("end_offset", &self.end_offset)
            .field("angle", &self.angle)
            .field("auto_rotate_angle", &self.auto_rotate_angle)
            .field("line_reduction", &self.line_reduction)
            .field("loop_count", &self.loop_count)
            .field("loop_distance", &self.loop_distance)
            .field("contour_first", &self.contour_first)
            .field("count", &self.count)
            .finish()
    }
}

#[derive(BinRead, BinWrite, Diff, PartialEq)]
#[diff(attr(
    #[derive(Debug, PartialEq)]
))]
#[brw(magic(15u32))] // Number of fields in this struct
pub struct HatchSettings2 {
    pub count: U32,
    _unknown_1: Field,
    pub pen: U32,
    pub flags: FieldOf<ObjectFlags>,
    pub edge_offset: F64,
    pub line_spacing: F64,
    pub start_offset: F64,
    pub end_offset: F64,
    pub angle: F64,
    pub auto_rotate_angle: F64,
    pub line_reduction: F64,
    pub loop_distance: F64,
    pub loop_count: U32,
    _unknown_2: [Field; 2],
}

// Custom Debug implementation to only print known fields
impl Debug for HatchSettings2 {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("HatchSettings2")
            .field("count", &self.count)
            .field("pen", &self.pen)
            .field("flags", &self.flags)
            .field("edge_offset", &self.edge_offset)
            .field("line_spacing", &self.line_spacing)
            .field("start_offset", &self.start_offset)
            .field("end_offset", &self.end_offset)
            .field("angle", &self.angle)
            .field("auto_rotate_angle", &self.auto_rotate_angle)
            .field("line_reduction", &self.line_reduction)
            .field("loop_distance", &self.loop_distance)
            .field("loop_count", &self.loop_count)
            .finish()
    }
}

#[derive(BinRead, BinWrite, Diff, PartialEq)]
#[diff(attr(
    #[derive(Debug, PartialEq)]
))]
pub struct Hatch {
    pub core: ObjectCore,
    pub outline: ArrayOf<Object>,
    pub settings_1: HatchSettings1,
    pub settings_2: HatchSettings2,
    _unknown_1: ArrayOf<Field>,
    _unknown_2: u32,
    _unknown_3: u32,
    _unknown_4: u32,
    pub pen: U32,
    _unknown_5: [Field; 2],
    pub name: WString,
    _unknown_6: [Field; 13],
    pub lines: ArrayOf<HatchLine>,
}

// Custom Debug implementation to only print known fields
impl Debug for Hatch {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("Hatch")
            .field("core", &self.core)
            .field("outline", &self.outline)
            .field("settings_1", &self.settings_1)
            .field("settings_2", &self.settings_2)
            .field("pen", &self.pen)
            .field("name", &self.name)
            .field("lines", &self.lines)
            .finish()
    }
}
