use std::{fs::File, path::PathBuf, time::Instant};

use binrw::{BinRead, BinWriterExt};
use clap::Parser;
use clap_verbosity_flag::{InfoLevel, Verbosity};
use diff::Diff;
use ezcad::file::EzCadHeader;
use log::{info, trace};

use crate::config::{Config, Operations};

mod config;

#[derive(Debug, Parser)]
struct Cli {
    /// Input file to parse
    #[arg(short, long)]
    input: PathBuf,

    /// Output file to write to
    #[arg(short, long)]
    output: Option<PathBuf>,

    /// Configuration file
    #[arg(short, long)]
    config: Option<PathBuf>,

    #[command(flatten)]
    verbose: Verbosity<InfoLevel>,
}

fn main() {
    let cli: Cli = Cli::parse();

    env_logger::builder()
        .format_timestamp(None)
        .format_target(false)
        .filter_level(cli.verbose.log_level_filter())
        .init();

    let mut input: File = File::open(cli.input).expect("Failed to open input file");

    let time: Instant = Instant::now();
    let mut file: EzCadHeader =
        EzCadHeader::read_le(&mut input).expect("Failed to parse input file as EZCAD format");
    info!("Input decode time: {:?}", time.elapsed());

    let pen_0 = file.pens_offset.data.pens.get(0).unwrap();
    let pen_1 = file.pens_offset.data.pens.get(1).unwrap();
    info!("{:#?}", pen_0.diff(pen_1));
    // info!("{:?}", diff.color.value.);

    // Print info on pens with non-default settings
    for pen in file
        .pens_offset
        .data
        .pens
        .iter()
        .filter(|x| *x.use_default == 0)
    {
        trace!("{:#?}", pen);
    }

    // Print all objects
    for layer in file.layers_offset.iter() {
        for object in layer.objects.iter() {
            trace!("{:#?}", object);
        }
    }

    if let Some(config) = cli.config {
        info!("Processing config file {}", config.to_string_lossy());
        let config: String = std::fs::read_to_string(config).expect("Failed to open config file");
        let config: Config = serde_yaml::from_str(&config).expect("Failed to parse config file");

        // Patch pen settings
        config.ops.apply(&mut file.pens_offset.data.pens);
    }

    if let Some(output) = cli.output {
        let mut output = File::create(output).expect("Failed to open output file");
        let time: Instant = Instant::now();
        output
            .write_le(&file)
            .expect("Failed to serialize contents in EZCAD format");
        info!("Output encode time: {:?}", time.elapsed());
    }
}
